@@ -46,15 +46,17 @@ hash_t adm_users;
 
 /* log options, for sure the trickiest :) */
 /* no backlog at all */
-int conf_backlog;
-int conf_memlog;
-int conf_log;
+int conf_backlog = 0;
+extern int conf_memlog;
+int conf_log = 0;
 /* number of lines in backlog */
 int conf_backlog_lines = 10;
+int conf_backlog_no_timestamp = 0;
 /* backlog even lines already backlogged */
-int conf_always_backlog;
-int conf_log_sync_interval;
+int conf_always_backlog = 0;
+int conf_log_sync_interval = 0;
 int conf_blreset_on_talk = 0;
+int conf_bl_msg_only = 0;
 
 list_t *parse_conf(FILE *file);
 static void conf_die(char *fmt, ...);
@@ -162,11 +164,22 @@ static pid_t daemonize(void)
 	case 0:
 		break;
 	default:
-		exit(0);
+		_exit(0);
 	}
+
 	if (setsid() < 0)
 		fatal("setsid() failed");
 
+	switch (fork()) {
+	case -1:
+		fatal("Fork failed");
+		break;
+	case 0:
+		break;
+	default:
+		_exit(0);
+	}
+
 	if (conf_log) {
 		snprintf(buf, 4095, "%s/bip.log", conf_log_root);
 		FILE *f = fopen(buf, "a");
@@ -395,7 +408,11 @@ void c_connection_free(struct c_connection *c)
 	}
 
 	free(c->away_nick);
-	free(c->on_connect_send);
+	free(c->no_client_away_msg);
+
+	char *s;
+	while ((s = list_remove_first(&c->on_connect_send)))
+		free(s);
 }
 
 static int add_connection(list_t *connectionl, list_t *data,
@@ -407,6 +424,7 @@ static int add_connection(list_t *connectionl, list_t *data,
 
 	c = calloc(sizeof(struct c_connection), 1);
 	list_init(&c->channell, NULL);
+	list_init(&c->on_connect_send, NULL);
 
 	while ((t = list_remove_first(data))) {
 		switch (t->type) {
@@ -454,8 +472,11 @@ static int add_connection(list_t *connectionl, list_t *data,
 		case LEX_AWAY_NICK:
 			c->away_nick = t->pdata;
 			break;
+		case LEX_NO_CLIENT_AWAY_MSG:
+			c->no_client_away_msg = t->pdata;
+			break;
 		case LEX_ON_CONNECT_SEND:
-			c->on_connect_send = t->pdata;
+			list_add_last(&c->on_connect_send, t->pdata);
 			break;
 		default:
 			conf_die("unknown keyword in connection statement");
@@ -595,12 +616,18 @@ int fireup(FILE *conf)
 		case LEX_BACKLOG:
 			conf_backlog = t->ndata;
 			break;
+		case LEX_BL_MSG_ONLY:
+			conf_bl_msg_only = t->ndata;
+			break;
 		case LEX_LOG:
 			conf_log = t->ndata;
 			break;
 		case LEX_BACKLOG_LINES:
 			conf_backlog_lines = t->ndata;
 			break;
+		case LEX_BACKLOG_NO_TIMESTAMP:
+			conf_backlog_no_timestamp = t->ndata;
+			break;
 		case LEX_LOG_ROOT:
 			if (conf_log_root)
 				free(conf_log_root);
@@ -703,16 +730,14 @@ int fireup(FILE *conf)
 #ifdef HAVE_LIBSSL
 	conf_ssl_certfile = NULL;	/* Make into a config option */
 	if (!conf_ssl_certfile) {
-		char *home = getenv("HOME");
-		char *ap = "/.bip/bip.pem";
-		if (!home)
-			fatal("no $HOME !, do you live in a trailer ?");
+		char *ap = "/bip.pem";
 		if (conf_ssl_certfile) {
 			free(conf_ssl_certfile);
 			conf_ssl_certfile = NULL;
 		}
-		conf_ssl_certfile = malloc(strlen(home) + strlen(ap) + 1);
-		strcpy(conf_ssl_certfile, home);
+		conf_ssl_certfile = malloc(strlen(conf_biphome) +
+				strlen(ap) + 1);
+		strcpy(conf_ssl_certfile, conf_biphome);
 		strcat(conf_ssl_certfile, ap);
 		mylog(LOG_INFO, "Default SSL certificate file: %s",
 				conf_ssl_certfile);
@@ -760,7 +785,7 @@ void ircize(list_t *ll)
 
 			link = hash_get(adm_conn, c->name);
 			if (!link) {
-				mylog(LOG_DEBUG, "new connection: \"%s\"",
+				mylog(LOG_DEBUGVERB, "new connection: \"%s\"",
 						c->name);
 				link = irc_link_new();
 				hash_insert(adm_conn, c->name, link);
@@ -783,7 +808,7 @@ void ircize(list_t *ll)
 				}
 				list_add_last(ll, link);
 			} else {
-				mylog(LOG_DEBUG, "old connection: \"%s\"",
+				mylog(LOG_DEBUGVERB, "old connection: \"%s\"",
 						c->name);
 #define MAYFREE(a) do { \
 		if (a) { \
@@ -813,9 +838,20 @@ void ircize(list_t *ll)
 
 			link->follow_nick = c->follow_nick;
 			link->ignore_first_nick = c->ignore_first_nick;
-			link->on_connect_send = strmaydup(c->on_connect_send);
+
+			char *s;
+			while ((s = list_remove_first(
+						&link->on_connect_send))) {
+				free(s);
+			}
+			list_append(&c->on_connect_send,
+					&link->on_connect_send);
+
 			link->away_nick = strmaydup(c->away_nick);
 
+			link->no_client_away_msg =
+				strmaydup(c->no_client_away_msg);
+
 			link->username = strmaydup(u->name);
 			link->password = malloc(20);
 			memcpy(link->password, u->password, 20);
@@ -904,7 +940,7 @@ int main(int argc, char **argv)
 	conf_global_log_file = stderr;
 	conf_pid_file = NULL;
 
-	while ((ch = getopt(argc, argv, "hnf:")) != -1) {
+	while ((ch = getopt(argc, argv, "hnf:s:")) != -1) {
 		switch (ch) {
 		case 'f':
 			confpath = strdup(optarg);
@@ -912,10 +948,16 @@ int main(int argc, char **argv)
 		case 'n':
 			conf_daemonize = 0;
 			break;
+		case 's':
+			conf_biphome = strdup(optarg);
+			break;
 		default:
 			usage(argv[0]);
 		}
 	}
+
+	umask(0027);
+
 	if (confpath) {
 		conf = fopen(confpath, "r");
 		if (!conf)
@@ -981,21 +1023,26 @@ void write_user_list(connection_t *c, char *dest)
 {
 	hash_iterator_t it;
 	list_iterator_t lit;
-	
+	char buf[4096];
+
 	WRITE_LINE2(c, P_IRCMASK, "PRIVMSG", dest, "bip user list:");
 	for (hash_it_init(&conf_users, &it); hash_it_item(&it);
 			hash_it_next(&it)) {
 		struct c_user *u = hash_it_item(&it);
 
-		WRITE_LINE2(c, P_IRCMASK, "PRIVMSG", dest, u->name);
+		snprintf(buf, 4095, "* %s:", u->name);
+		buf[4095] = 0;
+		WRITE_LINE2(c, P_IRCMASK, "PRIVMSG", dest, buf);
 		for (list_it_init(&u->connectionl, &lit); list_it_item(&lit);
 				list_it_next(&lit)) {
 			struct c_connection *con = list_it_item(&lit);
-			WRITE_LINE2(c, P_IRCMASK, "PRIVMSG", dest, con->name);
+			snprintf(buf, 4095, "  - %s", con->name);
+			buf[4095] = 0;
+			WRITE_LINE2(c, P_IRCMASK, "PRIVMSG", dest, buf);
 		}
 	}
 	WRITE_LINE2(c, P_IRCMASK, "PRIVMSG", dest,
-			"end of bip user list");
+			"End of bip user list");
 }
 
 #ifdef HAVE_LIBSSL
@@ -1026,7 +1073,7 @@ int ssl_check_trust(struct link_client *ic)
 	if(!LINK(ic)->untrusted_certs ||
 			sk_X509_num(LINK(ic)->untrusted_certs) <= 0)
 		return 0;
-	
+
 	trustcert = sk_X509_value(LINK(ic)->untrusted_certs, 0);
 	strcpy(subject, "Subject: ");
 	strcpy(issuer, "Issuer:  ");
@@ -1043,7 +1090,7 @@ int ssl_check_trust(struct link_client *ic)
 			"This server SSL certificate was not "
 			"accepted because it is not in your store "
 			"of trusted certificates:");
-	
+
 	WRITE_LINE2(CONN(ic), P_SERV, "NOTICE", "TrustEm", subject);
 	WRITE_LINE2(CONN(ic), P_SERV, "NOTICE", "TrustEm", issuer);
 	WRITE_LINE2(CONN(ic), P_SERV, "NOTICE", "TrustEm", fpstr);
@@ -1177,12 +1224,15 @@ void adm_ignore_first_nick(struct link_client *ic, char *val)
 void adm_on_connect_send(struct link_client *ic, char *val)
 {
 	struct link *link = LINK(ic);
-	if (link->on_connect_send) {
-		free(link->on_connect_send);
-		link->on_connect_send = NULL;
-	}
+	char *s;
+
 	if (val != NULL)
-		link->on_connect_send = strdup(val);
+		list_add_last(&link->on_connect_send, strdup(val));
+	else {
+		s = list_remove_last(&link->on_connect_send);
+		if (s)
+			free(s);
+	}
 }
 
 void adm_away_nick(struct link_client *ic, char *val)
@@ -1196,57 +1246,70 @@ void adm_away_nick(struct link_client *ic, char *val)
 		link->away_nick = strdup(val);
 }
 
-int adm_bip(struct link_client *ic, struct line *line);
-int adm_bip(struct link_client *ic, struct line *line)
+int adm_bip(struct link_client *ic, struct line *line, unsigned int privmsg)
 {
 	char *nick;
+
 	if (LINK(ic)->l_server)
 		nick = LINK(ic)->l_server->nick;
 	else
 		nick = LINK(ic)->prev_nick;
-	if (line->elemc < 2)
+	if (line->elemc < privmsg + 2)
 		return OK_FORGET;
 
-	if (strcasecmp(line->elemv[1], "RELOAD") == 0) {
+	if (strcasecmp(line->elemv[privmsg + 1], "RELOAD") == 0) {
 		reloading_client = ic;
 		sighup = 1;
-	} else if (strcasecmp(line->elemv[1], "LIST") == 0) {
+	} else if (strcasecmp(line->elemv[privmsg + 1], "LIST") == 0) {
 		write_user_list(CONN(ic), nick);
-	} else if (strcasecmp(line->elemv[1], "JUMP") == 0) {
+	} else if (strcasecmp(line->elemv[privmsg + 1], "JUMP") == 0) {
 		if (LINK(ic)->l_server) {
 			WRITE_LINE1(CONN(LINK(ic)->l_server), NULL, "QUIT",
 					"jumpin' jumpin'");
 			connection_close(CONN(LINK(ic)->l_server));
 		}
-	} else if (strcasecmp(line->elemv[1], "BLRESET") == 0) {
+	} else if (strcasecmp(line->elemv[privmsg + 1], "BLRESET") == 0) {
 		adm_blreset(ic);
-	} else if (strcasecmp(line->elemv[1], "HELP") == 0) {
+	} else if (strcasecmp(line->elemv[privmsg + 1], "HELP") == 0) {
 		WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG", nick,
-			"/BIP (RELOAD|LIST|JUMP|BLRESET|HELP)");
-	} else if (strcasecmp(line->elemv[1], "FOLLOW_NICK") == 0) {
-		if (line->elemc != 3)
+			"/BIP (RELOAD|LIST|JUMP|BLRESET|HELP|TRUST)");
+		WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG", nick,
+			"/BIP FOLLOW_NICK|IGNORE_FIRST_NICK TRUE|FALSE");
+		WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG", nick,
+			"/BIP ON_CONNECT_SEND <str> # Adds a string to "
+			"send on connect");
+		WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG", nick,
+			"/BIP ON_CONNECT_SEND # Clears on_connect_send");
+		WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG", nick,
+			"/BIP AWAY_NICK <nick> # Set away nick");
+		WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG", nick,
+			"/BIP AWAY_NICK # clear away nick");
+	} else if (strcasecmp(line->elemv[privmsg + 1], "FOLLOW_NICK") == 0) {
+		if (line->elemc != privmsg + 3)
 			return OK_FORGET;
-		adm_follow_nick(ic, line->elemv[2]);
-	} else if (strcasecmp(line->elemv[1], "IGNORE_FIRST_NICK") == 0) {
-		if (line->elemc != 3)
+		adm_follow_nick(ic, line->elemv[privmsg + 2]);
+	} else if (strcasecmp(line->elemv[privmsg + 1],
+				"IGNORE_FIRST_NICK") == 0) {
+		if (line->elemc != privmsg + 3)
 			return OK_FORGET;
-		adm_ignore_first_nick(ic, line->elemv[2]);
-	} else if (strcasecmp(line->elemv[1], "ON_CONNECT_SEND") == 0) {
-		if (line->elemc == 2)
+		adm_ignore_first_nick(ic, line->elemv[privmsg + 2]);
+	} else if (strcasecmp(line->elemv[privmsg + 1],
+				"ON_CONNECT_SEND") == 0) {
+		if (line->elemc == privmsg + 2)
 			adm_on_connect_send(ic, NULL);
-		else if (line->elemc == 3)
-			adm_on_connect_send(ic, line->elemv[2]);
+		else if (line->elemc == privmsg + 3)
+			adm_on_connect_send(ic, line->elemv[privmsg + 2]);
 		else
 			return OK_FORGET;
-	} else if (strcasecmp(line->elemv[1], "AWAY_NICK") == 0) {
-		if (line->elemc == 2)
+	} else if (strcasecmp(line->elemv[privmsg + 1], "AWAY_NICK") == 0) {
+		if (line->elemc == privmsg + 2)
 			adm_away_nick(ic, NULL);
-		else if (line->elemc == 3)
-			adm_away_nick(ic, line->elemv[2]);
+		else if (line->elemc == privmsg + 3)
+			adm_away_nick(ic, line->elemv[privmsg + 2]);
 		else
 			return OK_FORGET;
 #ifdef HAVE_LIBSSL
-	} else if (strcasecmp(line->elemv[1], "TRUST") == 0) {
+	} else if (strcasecmp(line->elemv[privmsg + 1], "TRUST") == 0) {
 		return adm_trust(ic, line);
 #endif
 	}
