@@ -40,10 +40,12 @@ static int irc_332(struct link_server *server, struct line *line);
 static int irc_333(struct link_server *server, struct line *line);
 static int irc_353(struct link_server *server, struct line *line);
 static int irc_366(struct link_server *server, struct line *line);
+static int irc_367(struct link_server *server, struct line *l);
+static int irc_368(struct link_server *server, struct line *l);
 void irc_server_shutdown(struct link_server *s);
+static int origin_is_me(struct line *l, struct link_server *server);
 
 void irc_client_free(struct link_client *cli);
-static char *make_irc_mask(char *nick, char *oldim);
 extern int conf_backlog;
 extern int conf_log_sync_interval;
 extern int conf_error;
@@ -74,26 +76,6 @@ struct channel *channel_new(const char *name)
 	return chan;
 }
 
-static char *make_irc_mask(char *nick, char *oldim)
-{
-	char *nm;
-	char *imaskend = oldim;
-	if (!imaskend) {
-		imaskend = "!bip@bip.bip.bip";
-	} else {
-		while (*imaskend && *imaskend != '!')
-			imaskend++;
-		if (*imaskend == '\0')
-			imaskend = "!bip@bip.bip.bip";
-	}
-
-	nm = malloc(strlen(nick) + strlen(imaskend) + 1);
-	*nm = 0;
-	strcat(nm, nick);
-	strcat(nm, imaskend);
-	return nm;
-}
-
 char *nick_from_ircmask(char *mask)
 {
 	char *nick = mask;
@@ -116,28 +98,6 @@ char *nick_from_ircmask(char *mask)
 	return ret;
 }
 
-static int is_irc_mask(char *mask)
-{
-	while (*mask && *mask != '!')
-		mask++;
-	if (*mask == 0)
-		return 0;
-	while (*mask && *mask != '@')
-		mask++;
-	if (*mask == 0)
-		return 0;
-	return 1;
-}
-
-static void set_ircmask(struct link_server *server, char *ircmask)
-{
-	if (!is_irc_mask(ircmask))
-		return;
-	if (server->irc_mask)
-		free(server->irc_mask);
-	server->irc_mask = strdup(ircmask);
-}
-
 #define NAMESIZE 256
 
 list_t *channel_name_list(struct channel *c)
@@ -186,68 +146,21 @@ list_t *channel_name_list(struct channel *c)
 
 static int irc_001(struct link_server *server, struct line *line)
 {
-	int ret = OK_FORGET;
-	char *p, *ps, *pe;
-	size_t len;
-	char *ircmask, *oldircmask;
+	(void)line;
 
 	if (LINK(server)->s_state == IRCS_WAS_CONNECTED)
 		LINK(server)->s_state = IRCS_RECONNECTING;
 	else
 		LINK(server)->s_state = IRCS_CONNECTING;
 
-	if (line->elemc != 3)
-		return ERR_PROTOCOL;
-
-	ps = pe = line->elemv[2];
-	pe += strlen(pe);
-	p = strrchr(ps, ' ');
-	if (!p)
-		return ERR_PROTOCOL;
-	p++;
-
-	len = pe - p;
-	ircmask = malloc(len + 1);
-	if (!ircmask)
-		fatal("malloc");
-	memcpy(ircmask, p, len);
-	ircmask[len] = 0;
-
-	oldircmask = server->irc_mask;
-	server->irc_mask = NULL;
-	set_ircmask(server, ircmask);
-	if (server->nick)
-		free(server->nick);
-	server->nick = nick_from_ircmask(ircmask);
-	if (server->irc_mask == NULL)
-		server->irc_mask = make_irc_mask(server->nick, "");
-
-	/* 001 means the nick we sent is accepted by server, inform clients */
-	if (oldircmask && LINK(server)->s_state == IRCS_RECONNECTING &&
-			LINK(server)->prev_nick) {
-		if (strcmp(LINK(server)->prev_nick, server->nick) != 0) {
-			/*
-			 * Inform connected clients that server<->proxy
-			 * disconnection resulted in nick change
-			 * sadly the oldircmask info is a guess only
-			 */
-			int i;
-			for (i = 0; i < LINK(server)->l_clientc; i++) {
-				struct link_client *c =
-					LINK(server)->l_clientv[i];
-				WRITE_LINE1(CONN(c), oldircmask, "NICK",
-						server->nick);
-			}
-		}
+	/* change nick on client */
+	int i;
+	for (i = 0; i < LINK(server)->l_clientc; i++) {
+		struct link_client *c = LINK(server)->l_clientv[i];
+		WRITE_LINE1(CONN(c), LINK(server)->cli_nick, "NICK",
+				server->nick);
 	}
-	if (LINK(server)->prev_nick)
-		free(LINK(server)->prev_nick);
-	LINK(server)->prev_nick = strdup(server->nick);
-
-	if (oldircmask)
-		free(oldircmask);
-	free(ircmask);
-	return ret;
+	return OK_COPY;
 }
 
 void irc_start_lagtest(struct link_server *l)
@@ -300,29 +213,44 @@ static void irc_server_join(struct link_server *s)
 static void irc_server_connected(struct link_server *server)
 {
 	int i;
-	char *initmask;
         LINK(server)->s_state = IRCS_CONNECTED;
+        LINK(server)->s_conn_attempt = 0;
+
+	mylog(LOG_INFO, "Connected user %s to %s", LINK(server)->username,
+		LINK(server)->name);
+
         irc_server_join(server);
         log_connected(LINK(server)->log);
 
-	/* Tell already connected clients what the nick is */
-	initmask = make_irc_mask(server->nick, server->irc_mask);
-
-	/* we change nick on client */
-	for (i = 0; i < LINK(server)->l_clientc; i++) {
-		struct link_client *ic = LINK(server)->l_clientv[i];
-		WRITE_LINE1(CONN(ic), initmask, "NICK", server->nick);
+	if (LINK(server)->cli_nick) {
+		/* we change nick on client */
+		for (i = 0; i < LINK(server)->l_clientc; i++) {
+			struct link_client *ic = LINK(server)->l_clientv[i];
+			WRITE_LINE1(CONN(ic), LINK(server)->cli_nick, "NICK",
+					server->nick);
+		}
+		free(LINK(server)->cli_nick);
+		LINK(server)->cli_nick = NULL;
 	}
-	free(initmask);
-
 	/* basic helper for nickserv and co */
-        if (LINK(server)->on_connect_send) {
-                ssize_t len = strlen(LINK(server)->on_connect_send) + 2;
+	list_iterator_t itocs;
+	for (list_it_init(&LINK(server)->on_connect_send, &itocs);
+				list_it_item(&itocs); list_it_next(&itocs)) {
+                ssize_t len = strlen(list_it_item(&itocs)) + 2;
                 char *str = malloc(len + 1);
-                sprintf(str, "%s\r\n", LINK(server)->on_connect_send);
+                sprintf(str, "%s\r\n", (char *)list_it_item(&itocs));
                 write_line(CONN(server), str);
                 free(str);
         }
+
+	if (LINK(server)->l_clientc == 0) {
+		if (LINK(server)->away_nick)
+			WRITE_LINE1(CONN(server), NULL, "NICK",
+					LINK(server)->away_nick);
+		if (LINK(server)->no_client_away_msg)
+			WRITE_LINE1(CONN(server), NULL, "AWAY",
+					LINK(server)->no_client_away_msg);
+	}
 }
 
 static int who_arg_to_ovmask(char *str)
@@ -336,46 +264,31 @@ static int who_arg_to_ovmask(char *str)
 }
 
 /*
- * Given the way irc nets disrespect the rfc, maybe we should completely forget
+ * Given the way irc nets disrespect the rfc, we completely forget
  * about this damn ircmask...
 :irc.iiens.net 352 pwet * ~a je.suis.t1r.net irc.iiens.net pwet H :0 d
 */
-static int parse_352(struct link_server *server, struct line *line)
+static int irc_352(struct link_server *server, struct line *line)
 {
-	char *im;
-
+	(void)server;
 	if (line->elemc < 8)
 		return ERR_PROTOCOL;
 
-	if (strcmp(line->elemv[2], "*") == 0) {
-		if (server->irc_mask &&
-				strcmp(line->elemv[1], server->nick) == 0) {
-			im = malloc(strlen(server->nick) +
-					strlen(line->elemv[3]) +
-					strlen(line->elemv[4]) + 2 + 1);
-			strcpy(im, server->nick);
-			strcat(im, "!");
-			strcat(im, line->elemv[3]);
-			strcat(im, "@");
-			strcat(im, line->elemv[4]);
-			set_ircmask(server, im);
-			free(im);
-			return OK_FORGET;
-		}
-	} else {
+	if (!origin_is_me(line, server)) {
 		struct channel *channel;
 		struct nick *nick;
 
 		channel = hash_get(&server->channels, line->elemv[2]);
 		if (!channel)
-			return OK_COPY;
+			return OK_COPY_WHO;
 
 		nick = hash_get(&channel->nicks, line->elemv[6]);
 		if (!nick)
-			return OK_COPY;
+			return OK_COPY_WHO;
 
 		nick->ovmask = who_arg_to_ovmask(line->elemv[7]);
 	}
+
 	return OK_COPY_WHO;
 }
 
@@ -383,16 +296,44 @@ static int irc_315(struct link_server *server, struct line *l)
 {
 	struct link *link = LINK(server);
 	if (link->who_client) {
-		--link->who_client->who_count;
-
-		if (link->who_client->who_count < 0)
-			fatal("negative who count");
+		if (link->who_client->who_count == 0) {
+			mylog(LOG_DEBUG, "Spurious irc_315");
+			return OK_COPY_WHO;
+		}
+		link->who_client->whoc_tstamp = time(NULL);
+		if (link->who_client->who_count > 0) {
+			--link->who_client->who_count;
+			mylog(LOG_DEBUG,
+				"RPL_ENDOFWHO: "
+				"Decrementing who count for %p: %d",
+				link->who_client, link->who_client->who_count);
+		}
 	}
 	l = NULL; /* keep gcc happy */
 
 	return OK_COPY_WHO;
 }
 
+void rotate_who_client(struct link *link)
+{
+	int i;
+	mylog(LOG_DEBUG, "rotate_who_client %p", link->who_client);
+	/* find a client with non-null who_count */
+	link->who_client = NULL;
+	for (i = 0; i < link->l_clientc; i++) {
+		struct link_client *ic = link->l_clientv[i];
+		if (!list_is_empty(&ic->who_queue)) {
+			char *l;
+			while ((l = list_remove_first(&ic->who_queue))) {
+				write_line(CONN(link->l_server), l);
+				free(l);
+			}
+			link->who_client = ic;
+			break;
+		}
+	}
+}
+
 /*
  * parses:join part mode kick kill privmsg quit nick names
  * returns: -1 invalid protocol
@@ -418,7 +359,9 @@ int irc_dispatch_server(struct link_server *server, struct line *line)
 		free(resps);
 		ret = OK_FORGET;
 	} else if (strcmp(line->elemv[0], "PONG") == 0) {
-		if (line->elemc == 3  && strcmp(line->elemv[2], S_PING) == 0) {
+		/* not all server reply with PONG <servername> <our string>
+		 * so we blindly assume the PONG is ours. */
+		if (line->elemc == 3) {
 			if (server->laginit_ts != -1) {
 				irc_compute_lag(server);
 				irc_lag_init(server);
@@ -474,6 +417,7 @@ int irc_dispatch_server(struct link_server *server, struct line *line)
 		if (LINK(server)->s_state == IRCS_CONNECTING) {
 			if (!list_is_empty(&LINK(server)->init_strings))
 				return ERR_PROTOCOL;
+			/* update the irc mask */
 			list_add_last(&LINK(server)->init_strings,
 					irc_line_dup(line));
 		}
@@ -484,13 +428,17 @@ int irc_dispatch_server(struct link_server *server, struct line *line)
 	} else if (strcmp(line->elemv[0], "333") == 0) {
 		ret = irc_333(server, line);
 	} else if (strcmp(line->elemv[0], "352") == 0) {
-		ret = parse_352(server, line);
+		ret = irc_352(server, line);
 	} else if (strcmp(line->elemv[0], "315") == 0) {
 		ret = irc_315(server, line);
 	} else if (strcmp(line->elemv[0], "353") == 0) {
 		ret = irc_353(server, line);
 	} else if (strcmp(line->elemv[0], "366") == 0) {
 		ret = irc_366(server, line);
+	} else if (strcmp(line->elemv[0], "367") == 0) {
+		ret = irc_367(server, line);
+	} else if (strcmp(line->elemv[0], "368") == 0) {
+		ret = irc_368(server, line);
 	} else if (strcmp(line->elemv[0], "PART") == 0) {
 		ret = irc_part(server, line);
 	} else if (strcmp(line->elemv[0], "MODE") == 0) {
@@ -523,32 +471,12 @@ int irc_dispatch_server(struct link_server *server, struct line *line)
 		s = irc_line_to_string(line);
 		write_line(CONN(LINK(server)->who_client), s);
 		free(s);
-
-		if (LINK(server)->who_client->who_count == 0) {
-			int i;
-			for (i = 0; i < LINK(server)->l_clientc; i++) {
-				struct link_client *ic =
-						LINK(server)->l_clientv[i];
-				if (ic == LINK(server)->who_client) {
-					if (!list_is_empty(&ic->who_queue))
-						fatal("who active and queued?");
-					continue;
-				}
-
-				if (!list_is_empty(&ic->who_queue)) {
-					char *l;
-					while ((l = list_remove_first(
-							&ic->who_queue))) {
-						write_line(CONN(server), l);
-						free(l);
-					}
-					LINK(server)->who_client = ic;
-					break;
-				}
-			}
-			if (i == LINK(server)->l_clientc)
-				LINK(server)->who_client = NULL;
-		}
+	}
+	if (LINK(server)->who_client &&
+			LINK(server)->who_client->who_count == 0) {
+		mylog(LOG_DEBUG, "OK_COPY_WHO: who_count for %p is nul",
+			LINK(server)->who_client);
+		rotate_who_client(LINK(server));
 	}
 	return ret;
 }
@@ -556,7 +484,14 @@ int irc_dispatch_server(struct link_server *server, struct line *line)
 /* send join and related stuff to client */
 static void irc_send_join(struct link_client *ic, struct channel *chan)
 {
-	WRITE_LINE1(CONN(ic), LINK(ic)->l_server->irc_mask, "JOIN", chan->name);
+	char *ircmask; /* fake an irc mask for rbot */
+
+	ircmask = malloc(strlen(LINK(ic)->l_server->nick) +
+			strlen("!bip@bip.bip.bip") + 1);
+	strcpy(ircmask, LINK(ic)->l_server->nick);
+	strcat(ircmask, "!bip@bip.bip.bip");
+	WRITE_LINE1(CONN(ic), ircmask, "JOIN", chan->name);
+	free(ircmask);
 	if (chan->topic)
 		WRITE_LINE3(CONN(ic), P_SERV, "332", LINK(ic)->l_server->nick,
 				chan->name, chan->topic);
@@ -567,8 +502,11 @@ static void irc_send_join(struct link_client *ic, struct channel *chan)
 	/* XXX: could be more efficient */
 	if (conf_backlog && log_has_backlog(LINK(ic)->log, chan->name)) {
 		char *line;
-		while ((line = log_backread(LINK(ic)->log, chan->name))) {
-			write_line(CONN(ic), line);
+		int skip = 0;
+		while ((line =
+		    log_backread(LINK(ic)->log, chan->name, &skip))) {
+			if (!skip)
+				write_line(CONN(ic), line);
 			free(line);
 		}
 		WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG", chan->name,
@@ -629,8 +567,16 @@ void unbind_from_link(struct link_client *ic)
 	if (i == l->l_clientc)
 		fatal("unbind_from_link");
 
-	if (l->who_client == ic)
+	if (l->who_client == ic) {
+		mylog(LOG_DEBUG, "unbind_from_link:  %p: %d",
+				l->who_client, ic->who_count);
 		l->who_client = NULL;
+	} else {
+		mylog(LOG_DEBUG,
+			"unbind_from_link: nothing to do %p != %p: %d",
+				ic, l->who_client,
+				ic->who_count);
+	}
 
 	for (i = i + 1; i < l->l_clientc; i++)
 		l->l_clientv[i - 1] = l->l_clientv[i];
@@ -648,7 +594,7 @@ void unbind_from_link(struct link_client *ic)
 
 int irc_cli_bip(struct link_client *ic, struct line *line)
 {
-	return adm_bip(ic, line);
+	return adm_bip(ic, line, 0);
 }
 
 #define PASS_SEP ':'
@@ -712,8 +658,10 @@ static void irc_cli_make_join(struct link_client *ic)
 
 		/* backlog privates */
 		char *str;
-		while ((str = log_backread(LINK(ic)->log, S_PRIVATES))) {
-			write_line(CONN(ic), str);
+		int skip = 0;
+		while ((str = log_backread(LINK(ic)->log, S_PRIVATES, &skip))) {
+			if (!skip)
+				write_line(CONN(ic), str);
 			free(str);
 		}
 	}
@@ -722,7 +670,6 @@ static void irc_cli_make_join(struct link_client *ic)
 static int irc_cli_startup(struct link_client *ic, struct line *line,
 		list_t *linkl)
 {
-	char *initmask;
 	char *init_nick;
 	char *user, *pass, *connname;
 	(void)line;
@@ -796,28 +743,16 @@ static int irc_cli_startup(struct link_client *ic, struct line *line,
 		return OK_CLOSE;
 	}
 
-#if 0
-	TODO code this -- OR NOT!
-	if (!connection_check_source(ic, ic->client->src_ipl))
-		return ERR_AUTH;
-#endif
-
 	TYPE(ic) = IRC_TYPE_CLIENT;
 
-	if (LINK(ic)->l_server)
-		initmask = make_irc_mask(init_nick,
-				LINK(ic)->l_server->irc_mask);
-	else
-		initmask = make_irc_mask(init_nick, LINK(ic)->prev_ircmask);
-
 	for (list_it_init(&LINK(ic)->init_strings, &it);
 			list_it_item(&it); list_it_next(&it))
 		write_init_string(CONN(ic), list_it_item(&it), init_nick);
 
-	/* we change nick on client */
+	/* we change nick on server */
 	if (LINK(ic)->l_server) {
 		struct link_server *server = LINK(ic)->l_server;
-		WRITE_LINE1(CONN(ic), initmask, "NICK", server->nick);
+		WRITE_LINE1(CONN(ic), init_nick, "NICK", server->nick);
 
 		if (!LINK(ic)->ignore_first_nick)
 			WRITE_LINE1(CONN(server), NULL, "NICK", init_nick);
@@ -825,9 +760,12 @@ static int irc_cli_startup(struct link_client *ic, struct line *line,
 				strcmp(LINK(ic)->away_nick, server->nick) == 0)
 			WRITE_LINE1(CONN(server), NULL, "NICK",
 					LINK(server)->connect_nick);
+
+		/* change away status */
+		if (server && LINK(ic)->no_client_away_msg)
+			WRITE_LINE0(CONN(server), NULL, "AWAY");
 	}
 
-	free(initmask);
 	if (!LINK(ic)->l_server) {
 		free(init_nick);
 		return OK_FORGET;
@@ -904,8 +842,12 @@ static int irc_cli_quit(struct link_client *ic, struct line *line)
 
 static int irc_cli_privmsg(struct link_client *ic, struct line *line)
 {
-	log_cli_privmsg(LINK(ic)->log, LINK(ic)->l_server->nick,
+	if (line->elemc >= 3)
+		log_cli_privmsg(LINK(ic)->log, LINK(ic)->l_server->nick,
 				line->elemv[1], line->elemv[2]);
+	if (strcmp(line->elemv[1], "-bip") == 0)
+		return adm_bip(ic, line, 1);
+
 	if (conf_blreset_on_talk)
 		adm_blreset(ic);
 	return OK_COPY_CLI;
@@ -922,6 +864,12 @@ static int irc_cli_who(struct link_client *ic, struct line *line)
 {
 	struct link *l = LINK(ic);
 
+	++ic->who_count;
+	if (ic->who_count == 1)
+		ic->whoc_tstamp = time(NULL);
+	mylog(LOG_DEBUG, "cli_who: Incrementing who count for %p: %d",
+				ic, ic->who_count);
+
 	if (l->who_client && l->who_client != ic) {
 		list_add_first(&ic->who_queue, irc_line_to_string(line));
 		return OK_FORGET;
@@ -930,51 +878,44 @@ static int irc_cli_who(struct link_client *ic, struct line *line)
 	if (!l->who_client)
 		l->who_client = ic;
 
+	return OK_COPY;
+}
+
+static int irc_cli_mode(struct link_client *ic, struct line *line)
+{
+	struct link *l = LINK(ic);
+
+	if (line->elemc != 3)
+		return OK_COPY;
+
+	/* This is a wild guess and that sucks. */
+	if (strcmp(line->elemv[0], "MODE") != 0 ||
+			strchr(line->elemv[2], 'b') == NULL)
+		return OK_COPY;
+
 	++ic->who_count;
+	if (ic->who_count == 1)
+		ic->whoc_tstamp = time(NULL);
+	mylog(LOG_DEBUG, "cli_mode: Incrementing who count for %p: %d",
+				l->who_client, ic->who_count);
+
+	if (l->who_client && l->who_client != ic) {
+		list_add_first(&ic->who_queue, irc_line_to_string(line));
+		return OK_FORGET;
+	}
+
+	if (!l->who_client)
+		l->who_client = ic;
 
 	return OK_COPY;
 }
 
-#if 0
-static void unbind_to_server(struct link_client *ic)
-{
-	struct irc_server *s = ic->client->server;
-	int i;
-	for (i = 0; i < s->clientc; i++)
-		if (s->clientv[i] == ic)
-			break;
-	if (i == s->clientc)
-		fatal("unbind_to_server");
-	for (i = i + 1; i < s->clientc; i++)
-		s->clientv[i - 1] = s->clientv[i];
-
-	s->clientc--;
-	s->clientv = realloc(s->clientv, s->clientc *
-			sizeof(struct irc_client *));
-	if (s->clientc == 0) { /* realloc was equiv to free() */
-		s->clientv = NULL;
-		return;
-	}
-	if (!s->clientv)
-		fatal("realloc");
-}
-
-static void bind_to_server(struct irc_client *ic)
-{
-	struct irc_server *s = ic->client->server;
-	int i = s->clientc;
-	s->clientc++;
-	s->clientv = realloc(s->clientv, s->clientc *
-			sizeof(struct irc_client *));
-	if (!s->clientv)
-		fatal("realloc");
-	s->clientv[i] = ic;
-}
-#endif
 
 static void irc_notify_disconnection(struct link_server *is)
 {
 	int i;
+	LINK(is)->cli_nick = strdup(is->nick);
+
 	for (i = 0; i < LINK(is)->l_clientc; i++) {
 		struct link_client *ic = LINK(is)->l_clientv[i];
 		hash_iterator_t hi;
@@ -985,7 +926,7 @@ static void irc_notify_disconnection(struct link_server *is)
 					c->name, is->nick,
 					"Server disconnected, reconnecting");
 		}
-		WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG", is->irc_mask,
+		WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG", is->nick,
 					"Server disconnected, reconnecting");
 	}
 }
@@ -1086,8 +1027,8 @@ static int irc_dispatch_trust_client(struct link_client *ic, struct line *line)
 	if (line->elemc < 2)
 		return ERR_PROTOCOL;
 
-	if (strcmp(line->elemv[0], "BIP") == 0 &&
-	    strcmp(line->elemv[1], "TRUST") == 0)
+	if (strcasecmp(line->elemv[0], "BIP") == 0 &&
+	    strcasecmp(line->elemv[1], "TRUST") == 0)
 		r = adm_trust(ic, line);
 
 	return r;
@@ -1127,16 +1068,18 @@ static int irc_dispatch_client(struct link_client *ic, struct line *line)
 		r = irc_cli_notice(ic, line);
 	} else if (strcmp(line->elemv[0], "WHO") == 0) {
 		r = irc_cli_who(ic, line);
+	} else if (strcmp(line->elemv[0], "MODE") == 0) {
+		r = irc_cli_mode(ic, line);
 	}
 
 	if (r == OK_COPY || r == OK_COPY_CLI) {
 		char *str = irc_line_to_string(line);
 		if (LINK(ic)->s_state == IRCS_CONNECTED &&
-				LINK(ic)->l_server->irc_mask)
+				LINK(ic)->l_server->nick)
 			write_line(CONN(LINK(ic)->l_server), str);
-		else if (LINK(ic)->l_server->irc_mask)
+		else if (LINK(ic)->l_server->nick)
 			WRITE_LINE2(CONN(ic), P_IRCMASK, "PRIVMSG",
-					LINK(ic)->l_server->irc_mask,
+					LINK(ic)->l_server->nick,
 					":Not connected please try again "
 					"later...\r\n");
 
@@ -1167,7 +1110,7 @@ static void irc_copy_cli(struct link_client *src, struct link_client *dest,
 	}
 
 	if (ischannel(*line->elemv[1]) || LINK(src) != LINK(dest)) {
-		line->origin = LINK(src)->l_server->irc_mask;
+		line->origin = LINK(src)->l_server->nick;
 		str = irc_line_to_string(line);
 		line->origin = NULL;
 		write_line(CONN(dest), str);
@@ -1189,7 +1132,7 @@ static void irc_copy_cli(struct link_client *src, struct link_client *dest,
 
 	line->origin = line->elemv[1];
 	/* tricky: */
-	line->elemv[1] = LINK(src)->l_server->irc_mask;
+	line->elemv[1] = LINK(src)->l_server->nick;
 
 	free(line->elemv[2]);
 	line->elemv[2] = tmp;
@@ -1249,7 +1192,7 @@ static int origin_is_me(struct line *l, struct link_server *server)
 	if (!l->origin)
 		return 0;
 	nick = nick_from_ircmask(l->origin);
-	if (strcmp(nick, server->nick) == 0) {
+	if (strcasecmp(nick, server->nick) == 0) {
 		free(nick);
 		return 1;
 	}
@@ -1421,6 +1364,37 @@ static int irc_366(struct link_server *server, struct line *line)
 	return OK_COPY;
 }
 
+static int irc_367(struct link_server *server, struct line *l)
+{
+	(void)server;
+	(void)l;
+	return OK_COPY_WHO;
+}
+
+/* same as irc_315 */
+static int irc_368(struct link_server *server, struct line *l)
+{
+	struct link *link = LINK(server);
+	if (link->who_client) {
+		if (link->who_client->who_count == 0) {
+			mylog(LOG_DEBUG, "Spurious irc_368");
+			return OK_COPY_WHO;
+		}
+		link->who_client->whoc_tstamp = time(NULL);
+
+		if (link->who_client->who_count > 0) {
+			--link->who_client->who_count;
+			mylog(LOG_DEBUG,
+				"RPL_ENDOFBANLIST: "
+				"Decrementing who count for %p: %d",
+				link->who_client, link->who_client->who_count);
+		}
+	}
+	l = NULL; /* keep gcc happy */
+
+	return OK_COPY_WHO;
+}
+
 static void channel_free(struct channel *c)
 {
 	if (c->name)
@@ -1750,7 +1724,7 @@ static int irc_nick(struct link_server *server, struct line *line)
 	struct channel *channel;
 	struct nick *nick;
 	hash_iterator_t hi;
-	char *s_nick;
+	char *org_nick;
 
 	if (line->elemc != 2)
 		return ERR_PROTOCOL;
@@ -1758,19 +1732,25 @@ static int irc_nick(struct link_server *server, struct line *line)
 	if (!line->origin)
 		return ERR_PROTOCOL;
 
-	s_nick = nick_from_ircmask(line->origin);
+	org_nick = nick_from_ircmask(line->origin);
 
-	if (strcmp(s_nick, server->nick) == 0) {
-		char *nim;
+	for (hash_it_init(&server->channels, &hi); hash_it_item(&hi);
+			hash_it_next(&hi)) {
+		channel = hash_it_item(&hi);
+		nick = hash_get(&channel->nicks, org_nick);
+		if (!nick)
+			continue;
+		hash_remove(&channel->nicks, org_nick);
+		free(nick->name);
+		nick->name = strdup(line->elemv[1]);
+		hash_insert(&channel->nicks, nick->name, nick);
+		log_nick(LINK(server)->log, org_nick, channel->name,
+				line->elemv[1]);
+	}
+
+	if (origin_is_me(line, server)) {
 		free(server->nick);
 		server->nick = strdup(line->elemv[1]);
-
-		/* How the hell do we handle that crap cleanly ? */
-		nim = make_irc_mask(line->elemv[1], server->irc_mask);
-		if (nim) {
-			free(server->irc_mask);
-			server->irc_mask = nim;
-		}
 		if (LINK(server)->follow_nick &&
 				(LINK(server)->away_nick == NULL ||
 				strcmp(server->nick, LINK(server)->away_nick))
@@ -1780,21 +1760,7 @@ static int irc_nick(struct link_server *server, struct line *line)
 		}
 	}
 
-	for (hash_it_init(&server->channels, &hi); hash_it_item(&hi);
-			hash_it_next(&hi)) {
-		channel = hash_it_item(&hi);
-		nick = hash_get(&channel->nicks, s_nick);
-		if (!nick)
-			continue;
-		hash_remove(&channel->nicks, s_nick);
-		free(nick->name);
-		nick->name = strdup(line->elemv[1]);
-		hash_insert(&channel->nicks, nick->name, nick);
-		log_nick(LINK(server)->log, s_nick, channel->name,
-				line->elemv[1]);
-	}
-
-	free(s_nick);
+	free(org_nick);
 	return OK_COPY;
 }
 
@@ -1884,10 +1850,6 @@ void server_cleanup(struct link_server *server)
 		free(server->nick);
 		server->nick = NULL;
 	}
-	if (server->irc_mask) {
-		free(server->irc_mask);
-		server->irc_mask = NULL;
-	}
 	if (LINK(server)->s_state == IRCS_CONNECTED) {
 		LINK(server)->s_state = IRCS_WAS_CONNECTED;
 	} else {
@@ -1921,6 +1883,9 @@ void irc_client_close(struct link_client *ic)
 			if (is && LINK(ic)->away_nick)
 				WRITE_LINE1(CONN(is), NULL, "NICK",
 						LINK(ic)->away_nick);
+			if (is && LINK(ic)->no_client_away_msg)
+				WRITE_LINE1(CONN(is), NULL, "AWAY",
+						LINK(ic)->no_client_away_msg);
 			log_client_none_connected(LINK(ic)->log);
 		}
 		irc_client_free(ic);
@@ -1942,6 +1907,8 @@ static void irc_close(struct link_any *l)
 
 		if (LINK(is)->s_state == IRCS_CONNECTED)
 			irc_notify_disconnection(is);
+		else
+			LINK(is)->s_conn_attempt++;
 		irc_server_shutdown(is);
 		log_disconnected(LINK(is)->log);
 
@@ -1950,7 +1917,7 @@ static void irc_close(struct link_any *l)
 		if (LINK(is)->last_connection &&
 				time(NULL) - LINK(is)->last_connection
 					< CONN_INTERVAL)
-			timer = RECONN_TIMER;
+			timer = RECONN_TIMER * (LINK(is)->s_conn_attempt);
 		mylog(LOG_ERROR, "%s dead, reconnecting in %d seconds",
 				LINK(l)->name, timer);
 		LINK(is)->recon_timer = timer;
@@ -1995,8 +1962,6 @@ void irc_server_free(struct link_server *s)
 {
 	if (s->nick)
 		free(s->nick);
-	if (s->irc_mask)
-		free(s->irc_mask);
 	if (s->user_mode)
 		free(s->user_mode);
 	free(s);
@@ -2057,46 +2022,127 @@ void irc_server_shutdown(struct link_server *s)
 	if (LINK(s)->prev_nick)
 		free(LINK(s)->prev_nick);
 	LINK(s)->prev_nick = strdup(s->nick);
-
-	if (LINK(s)->prev_ircmask) {
-		free(LINK(s)->prev_ircmask);
-		LINK(s)->prev_ircmask = NULL;
-	}
-	if (s->irc_mask)
-		LINK(s)->prev_ircmask = strdup(s->irc_mask);
 }
 
 
 #ifdef HAVE_OIDENTD
-/* ugly */
+
+#define BIP_OIDENTD_START "## AUTOGENERATED BY BIP. DO NOT EDIT ##\n"
+#define BIP_OIDENTD_END "## END OF AUTOGENERATED STUFF ##\n"
+#define BIP_OIDENTD_END_LENGTH strlen(BIP_OIDENTD_END)
+
 void oidentd_dump(list_t *connl)
 {
 	list_iterator_t it;
 	FILE *f;
-	char *home;
-	char tmpbuf[256];
-	int err;
+	char *home, *filename;
+	char *bipstart = NULL, *bipend = NULL;
+	struct stat stats;
+	char tag_written = 0;
 
 	home = getenv("HOME");
-	if (!home)
-		return;
-	strcpy(tmpbuf, home);
-	strcat(tmpbuf, "/.oidentd.conf");
-	f = fopen(tmpbuf, "w");
-	if (!f)
+	if (home == NULL) {
+		mylog(LOG_WARN, "Can't get $HOME, not writing oidentd.conf");
 		return;
+	}
+
+	filename = (char *)malloc(strlen(home) + strlen("/.oidentd.conf"));
+	if(filename == NULL)
+		fatal("Out of memory.");
+
+	sprintf(filename, "%s/.oidentd.conf", home);
+
+	if (stat(filename, &stats) == -1) {
+		if (errno == ENOENT) {
+			f = fopen(filename, "w+");
+			fchmod(fileno(f), 0644);
+		} else {
+			mylog(LOG_WARN, "Can't open/create %s", filename);
+			free(filename);
+			return;
+		}
+	} else {
+		char *content;
+		f = fopen(filename, "r+");
+
+		if (!f) {
+			mylog(LOG_WARN, "Can't open/create %s", filename);
+			free(filename);
+			return;
+		}
+
+		content = (char *)malloc(stats.st_size + 1);
+
+		if (content == NULL){
+			mylog(LOG_WARN, "oidentd_dump : malloc failed, "
+					"returning");
+			fclose(f);
+			free(filename);
+			return;
+		}
+
+		if (fread(content, 1, stats.st_size, f) !=
+				(size_t)stats.st_size) {
+			mylog(LOG_WARN, "Can't read %s fully", filename);
+			free(content);
+			goto clean_oidentd;
+		}
+
+		/* Set terminating zero for strstr */
+		content[stats.st_size] = '\0';
+
+		bipstart = strstr(content, BIP_OIDENTD_START);
+		if (bipstart != NULL) {
+			/* We have some config left, rewrite the file
+			 * completely */
+			fseek(f, SEEK_SET, 0);
+			if (ftruncate(fileno(f), 0) == -1) {
+				mylog(LOG_DEBUG, "Can't reset %s size",
+						filename);
+				free(content);
+				goto clean_oidentd;
+			}
+
+			bipend = strstr(bipstart, BIP_OIDENTD_END);
+
+			/* data preceeding the tag */
+			fwrite(content, 1, bipstart - content, f);
+
+			/* data following the tag, if any */
+			if (bipend != NULL)
+				fwrite(bipend + BIP_OIDENTD_END_LENGTH, 1,
+						stats.st_size -
+						(bipend - content) -
+						BIP_OIDENTD_END_LENGTH, f);
+			else
+				mylog(LOG_WARN, "No %s mark found in %s",
+						BIP_OIDENTD_END, filename);
+		} else {
+			/* No previous conf */
+			if (stats.st_size != 0 &&
+					content[stats.st_size - 1] != '\n')
+				fprintf(f, "\n");
+		}
+	}
 
 	for (list_it_init(connl, &it); list_it_item(&it); list_it_next(&it)) {
 		connection_t *c = list_it_item(&it);
 		struct link_any *la = c->user_data;
 		if (c->connected == CONN_OK && la &&
 				TYPE(la) == IRC_TYPE_SERVER) {
-			struct link_server *ls = (struct link_server*)la;
-			struct link *l = LINK(ls);
-
+			struct link_server *ls;
+			struct link *l;
 			char *localip, *remoteip;
 			int localport, remoteport;
 
+			if (!tag_written) {
+				fprintf(f, BIP_OIDENTD_START);
+				tag_written = 1;
+			}
+
+			ls = (struct link_server*)la;
+			l = LINK(ls);
+
 			localip = connection_localip(CONN(ls));
 			localport = connection_localport(CONN(ls));
 			remoteip = connection_remoteip(CONN(ls));
@@ -2111,19 +2157,41 @@ void oidentd_dump(list_t *connl)
 			free(remoteip);
 		}
 	}
-	fflush(f);
+	if (tag_written)
+		fprintf(f, BIP_OIDENTD_END);
+
+clean_oidentd:
 	fclose(f);
-	err = chmod(tmpbuf, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
-	if (err) {
-		mylog(LOG_WARN, "Unable to set modes for '%s': %s", tmpbuf,
-				strerror(errno));
-	}
+	free(filename);
 }
 #endif
 
 /* do not use */
 list_t *_connections = NULL;
 
+void timeout_clean_who_counts(list_t *conns)
+{
+	list_iterator_t it;
+	for (list_it_init(conns, &it); list_it_item(&it); list_it_next(&it)) {
+		struct link *l = list_it_item(&it);
+		struct link_client *client = l->who_client;
+
+		if (client && client->whoc_tstamp) {
+			time_t now;
+			now = time(NULL);
+			if (now - client->whoc_tstamp > 10) {
+				mylog(LOG_DEBUG, "Yawn, "
+						"forgetting one who reply");
+				if (client->who_count > 0)
+					--client->who_count;
+				client->whoc_tstamp = time(NULL);
+				if (client->who_count == 0)
+					rotate_who_client(l);
+			}
+		}
+	}
+}
+
 struct link_client *reloading_client;
 /*
  * The main loop
@@ -2190,8 +2258,10 @@ void irc_main(connection_t *inc, list_t *ll)
 		struct link *link;
 		connection_t *conn;
 
-		/* Compute timeouts for next reconnections and lagouts */
 		if (timeleft == 0) {
+			/*
+			 * Compute timeouts for next reconnections and lagouts
+			 */
 			static int throttle_prot = S_CONN_DELAY - 1;
 
 			timeleft = 1000;
@@ -2250,6 +2320,11 @@ void irc_main(connection_t *inc, list_t *ll)
 					list_it_remove(&li);
 				}
 			}
+
+			/*
+			 * Cleanup lagging or dangling who_count buffers
+			 */
+			timeout_clean_who_counts(ll);
 		}
 
 		int nc;
@@ -2391,6 +2466,7 @@ struct link *irc_link_new()
 
 	hash_init(&link->chan_infos, HASH_NOCASE);
 	list_init(&link->chan_infos_order, list_ptr_cmp);
+	list_init(&link->on_connect_send, list_ptr_cmp);
 	return link;
 }
 
